# -*- coding: utf-8 -*-
"""
Auth* related model.

This is where the models used by the authentication stack are defined.

It's perfectly fine to re-use this definition in the italica application,
though.

"""
import os
from datetime import datetime
from hashlib import sha256
__all__ = ['User', 'Group', 'Permission']

import ming
from ming import schema as s
from ming import Session
from ming.odm import FieldProperty, ForeignIdProperty, RelationProperty
from ming.odm.declarative import MappedClass
from ming.odm import ThreadLocalODMSession

mainsession = Session()
DBSession = ThreadLocalODMSession(mainsession)


def init_model(engine):
    mainsession.bind = engine
    ming.odm.Mapper.compile_all()

    for mapper in ming.odm.Mapper.all_mappers():
        mainsession.ensure_indexes(mapper.collection)

    return DBSession


class Group(MappedClass):
    """
    Group definition.
    """
    class __mongometa__:
        session = DBSession
        name = 'tg_group'
        unique_indexes = [('group_name',),]

    _id = FieldProperty(s.ObjectId)
    group_name = FieldProperty(s.String)
    display_name = FieldProperty(s.String)

    permissions = RelationProperty('Permission')


class Permission(MappedClass):
    """
    Permission definition.
    """
    class __mongometa__:
        session = DBSession
        name = 'tg_permission'
        unique_indexes = [('permission_name',),]

    _id = FieldProperty(s.ObjectId)
    permission_name = FieldProperty(s.String)
    description = FieldProperty(s.String)

    _groups = ForeignIdProperty(Group, uselist=True)
    groups = RelationProperty(Group)


class User(MappedClass):
    """
    User definition.

    This is the user definition used by :mod:`repoze.who`, which requires at
    least the ``user_name`` column.

    """
    class __mongometa__:
        session = DBSession
        name = 'tg_user'
        unique_indexes = [('user_name',),]

    class PasswordProperty(FieldProperty):
        @classmethod
        def _hash_password(cls, password):
            salt = sha256()
            salt.update(os.urandom(60))
            salt = salt.hexdigest()

            hash = sha256()
            # Make sure password is a str because we cannot hash unicode objects
            hash.update((password + salt).encode('utf-8'))
            hash = hash.hexdigest()

            password = salt + hash


            return password

        def __set__(self, instance, value):
            value = self._hash_password(value)
            return FieldProperty.__set__(self, instance, value)

    _id = FieldProperty(s.ObjectId)
    user_name = FieldProperty(s.String)
    email_address = FieldProperty(s.String)
    display_name = FieldProperty(s.String)

    _groups = ForeignIdProperty(Group, uselist=True)
    groups = RelationProperty(Group)

    password = PasswordProperty(s.String)
    created = FieldProperty(s.DateTime, if_missing=datetime.now)

    @property
    def permissions(self):
        return Permission.query.find(dict(_groups={'$in':self._groups})).all()

    @classmethod
    def by_email_address(cls, email):
        """Return the user object whose email address is ``email``."""
        return cls.query.get(email_address=email)

    def validate_password(self, password):
        """
        Check the password against existing credentials.

        :param password: the password that was provided by the user to
            try and authenticate. This is the clear text version that we will
            need to match against the hashed one in the database.
        :type password: unicode object.
        :return: Whether the password is valid.
        :rtype: bool

        """
        hash = sha256()
        hash.update((password + self.password[:64]).encode('utf-8'))
        return self.password[64:] == hash.hexdigest()



# -*- coding: utf-8 -*-

from tg import config


def get_new_permission_form():
    new_permission_config = config['_pluggable_tgapppermissions_config']
    new_permission_form = new_permission_config.get('new_permission_form_instance')
    if not new_permission_form:
        form_path = new_permission_config.get('new_permission_form', 'tgapppermissions.lib.forms.NewPermission')
        module, form_name = form_path.rsplit('.', 1)
        module = __import__(module, fromlist=form_name)
        form_class = getattr(module, form_name)
        new_permission_form = new_permission_config['new_permission_form_instance'] = form_class()

    return new_permission_form


def get_edit_permission_form():
    edit_permission_config = config['_pluggable_tgapppermissions_config']
    edit_permission_form = edit_permission_config.get('edit_permission_form_instance')
    if not edit_permission_form:
        form_path = edit_permission_config.get('edit_permission_form', 'tgapppermissions.lib.forms.EditPermission')
        module, form_name = form_path.rsplit('.', 1)
        module = __import__(module, fromlist=form_name)
        form_class = getattr(module, form_name)
        edit_permission_form = edit_permission_config['edit_permission_form_instance'] = form_class()

    return edit_permission_form


def get_edit_user_form():
    edit_user_config = config['_pluggable_tgapppermissions_config']
    edit_user_form = edit_user_config.get('edit_user_form_instance')
    if not edit_user_form:
        form_path = edit_user_config.get('edit_user_form', 'tgapppermissions.lib.forms.EditUser')
        module, form_name = form_path.rsplit('.', 1)
        module = __import__(module, fromlist=form_name)
        form_class = getattr(module, form_name)
        edit_user_form = edit_user_config['edit_user_form_instance'] = form_class()

    return edit_user_form



# -*- coding: utf-8 -*-
"""Setup the tgapppermissions application"""
import logging
from tgapppermissions import model
from tgext.pluggable import app_model

log = logging.getLogger(__name__)

def bootstrap(command, conf, vars):
    log.info("bootstrapping tgapppermissions")
    p = app_model.Permission(permission_name='tgapppermissions',
                             description='Permits to manage permissions')
    try:
        model.DBSession.add(p)
    except AttributeError:
        # mute ming complaints
        pass
    model.DBSession.flush()


# -*- coding: utf-8 -*-
"""Main Controller"""

from tg import TGController, abort
from tg import expose, flash, require, url, lurl, request, redirect, validate, predicates
from tg.decorators import paginate
from tg.i18n import ugettext as _

from tgapppermissions import model
from tgext.pluggable import app_model, plug_url, plug_redirect

from tgapppermissions.lib import get_new_permission_form, get_edit_permission_form, get_edit_user_form
from tgapppermissions.helpers import get_primary_field


class RootController(TGController):
    allow_only = predicates.has_permission('tgapppermissions')

    @expose('tgapppermissions.templates.index')
    def index(self):
        count, permissions = model.provider.query(app_model.Permission)
        return dict(permissions_count=count,
                    permissions=permissions,
                    mount_point=self.mount_point)

    @expose('tgapppermissions.templates.new_permission')
    def new_permission(self, **_):
        return dict(form=get_new_permission_form(),
                    mount_point=self.mount_point,
                    action=plug_url('tgapppermissions', '/create_permission'),
                    values=None)

    @expose()
    @validate(get_new_permission_form(), error_handler=new_permission)
    def create_permission(self, **kwargs):
        dictionary = {
            'permission_name': kwargs.get('permission_name'),
            'description': kwargs.get('description'),
            'groups': kwargs.get('groups'),
        }
        model.provider.create(app_model.Permission, dictionary)

        flash(_('Permission created.'))
        return redirect(url(self.mount_point))

    @expose('tgapppermissions.templates.edit_permission')
    def edit_permission(self, permission_id, **_):
        primary_field = model.provider.get_primary_field(app_model.Permission)
        permission = model.provider.get_obj(app_model.Permission,
                                            {primary_field: permission_id}) or abort(404)
        values = model.provider.dictify(permission)
        values['groups'] = map(lambda x: str(x), permission['_groups'])
        return dict(form=get_edit_permission_form(),
                    mount_point=self.mount_point,
                    action=plug_url('tgapppermissions', '/update_permission/' + permission_id),
                    values=values)

    @expose()
    @validate(get_edit_permission_form(), error_handler=edit_permission)
    def update_permission(self, permission_id, **kwargs):
        primary_field = model.provider.get_primary_field(app_model.Permission)
        model.provider.update(app_model.Permission,
                              {primary_field: permission_id,
                               'permission_name': kwargs.get('permission_name'),
                               'description': kwargs.get('description'),
                               'groups': kwargs.get('groups')})
        flash(_('Permission updated.'))
        return redirect(url(self.mount_point))

    @expose()
    def delete_permission(self, permission_id):
        primary_field = model.provider.get_primary_field(app_model.Permission)
        try:
            model.provider.delete(app_model.Permission, {primary_field: permission_id})
        except AttributeError:
            abort(404)
        flash(_('Permission deleted'))
        return redirect(url(self.mount_point))

    @expose('tgapppermissions.templates.users')
    @paginate('users', items_per_page=20)
    def users(self, search_by=None, search_value=None):
        query_args = {}
        if search_by:
            query_args = dict(filters={search_by: search_value},
                              substring_filters=[search_by])
        _, users = model.provider.query(app_model.User, order_by='display_name', **query_args)
        return dict(mount_point=self.mount_point,
                    users=users)

    @expose('tgapppermissions.templates.edit_user')
    def edit_user(self, user_id, **_):
        primary_field = get_primary_field(app_model.User)
        user = model.provider.get(app_model.User, {primary_field: user_id}) or abort(404)
        return dict(
            form=get_edit_user_form(),
            mount_point=self.mount_point,
            action=plug_url('tgapppermissions', '/update_user/' + user_id),
            values=user,
        )

    @expose()
    @validate(get_edit_user_form(), error_handler=edit_user)
    def update_user(self, user_id, **kwargs):
        """currently updates ONLY the groups of the user"""
        primary_field = get_primary_field(app_model.User)
        model.provider.update(app_model.User,
                              {primary_field: user_id,
                               'groups': kwargs.get('groups')})
        flash(_('User updated.'))
        return redirect(url(self.mount_point + '/users'))



# -*- coding: utf-8 -*-

"""WebHelpers used in tgapp-permissions."""
from tgext.pluggable import app_model
from tgapppermissions import model
import six


def get_primary_field(_model):
    return model.provider.get_primary_field(
        model.provider.get_entity(_model)
    )
